@@ -1,9 +1,17 @@
+# -*- coding: utf-8 -*-
 from __future__ import division
+from __future__ import absolute_import
 
-import base64
 import os
 import time
-import cStringIO
+
+from aetros.backend import JobImage
+
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from io import StringIO
+from io import BytesIO
 
 import PIL.Image
 import math
@@ -12,20 +20,17 @@ from keras import backend as K
 import keras.layers.convolutional
 
 import numpy as np
-from keras.models import Sequential
 
 from aetros.utils.image import get_layer_vis_square
-from network import ensure_dir, get_total_params
-
+from .keras_model_utils import ensure_dir, get_total_params
+import six
 
 class KerasLogger(Callback):
-    def __init__(self, trainer, backend, job_model, general_logger):
+    def __init__(self, trainer, job_backend, general_logger):
         self.params = {}
         super(KerasLogger, self).__init__()
         self.validation_per_batch = []
         self.ins = None
-
-        self.trainer = None
         self.insights_sample_path = None
 
         self.trainer = trainer
@@ -33,8 +38,8 @@ class KerasLogger(Callback):
         self.log_epoch = True
         self.confusion_matrix = True
 
-        self.backend = backend
-        self.job_model = job_model
+        self.job_backend = job_backend
+        self.job_model = job_backend.get_job_model()
         self.general_logger = general_logger
         self._test_with_acc = None
         self.last_batch_time = time.time()
@@ -43,17 +48,17 @@ class KerasLogger(Callback):
         self.batches_per_second = 0
         self.stats = []
         self.last_current = None
-        self.filepath_best = job_model.get_weights_filepath_best()
-        self.filepath_latest = job_model.get_weights_filepath_latest()
+        self.filepath_best = self.job_model.get_weights_filepath_best()
+        self.filepath_latest = self.job_model.get_weights_filepath_latest()
 
         ensure_dir(os.path.dirname(self.filepath_best))
         self.data_gathered = False
 
-        self.insights_x = None
+        self.accuracy_channel = None
+        self.loss_channel = None
 
-        self.best_epoch = 0
+        self.insights_x = None
         self.best_total_accuracy = 0
-        self.worst_total_accuracy = 0
 
     def on_train_begin(self, logs={}):
         self.start_time = time.time()
@@ -64,16 +69,30 @@ class KerasLogger(Callback):
         self.current['epoch'] = 0
         self.current['started'] = self.start_time
         self.trainer.set_job_info('current', self.current)
-        nb_sample = self.params['nb_sample'] #training samples total
-        nb_epoch = self.params['nb_epoch'] #training epoches total
+        nb_sample = self.params['nb_sample']  # training samples total
+        nb_epoch = self.params['nb_epoch']  # training epoches total
 
         self.current['nb_sample'] = nb_sample
         self.current['nb_epoch'] = nb_epoch
 
+        xaxis = {
+            'range': [1, nb_epoch],
+            'title': u'Epoch ⇢'
+        }
+        yaxis = {
+            'range': [1, 100],
+            'title': u'% Accuracy ⇢',
+            'dtick': 10
+        }
+
+        self.accuracy_channel = self.job_backend.create_channel('accuracy', main_graph=True, xaxis=xaxis, yaxis=yaxis)
+        self.loss_channel = self.job_backend.create_loss_channel('loss', xaxis=xaxis)
+
     def on_batch_begin(self, batch, logs={}):
 
         if not self.data_gathered:
-            # we need to do it in on_batch_begin due to the fact that self.model.validation_data is not availabe in on_train_begin
+            # we need to do it in on_batch_begin due to the fact that
+            # self.model.validation_data is not availabe in on_train_begin
             self.data_gathered = True
             dataset_infos = {}
             dataset_info = {
@@ -84,7 +103,7 @@ class KerasLogger(Callback):
             self.trainer.set_job_info('datasets', dataset_infos)
 
         batch_size = logs['size']
-        nb_batches = math.ceil(self.current['nb_sample'] / batch_size) #normal nb batches
+        nb_batches = math.ceil(self.current['nb_sample'] / batch_size)  # normal nb batches
 
         self.current['nb_batches'] = nb_batches
         self.current['batch_size'] = batch_size
@@ -96,41 +115,37 @@ class KerasLogger(Callback):
         self.validation_per_batch.append(loss)
 
         current_batch = logs['batch']
-        current_batch_size = logs['size'] #how many training items in this batch, differs for the last run
+        # how many training items in this batch, differs for the last run
+        current_batch_size = logs['size']
 
         self.made_batches += 1
 
         time_diff = time.time() - self.last_batch_time
 
-        if time_diff > 1 or batch == self.current['nb_batches']: #only each second or last batch
+        if time_diff > 2 or batch == self.current['nb_batches']:  # only each second second or last batch
             self.batches_per_second = self.made_batches / time_diff
             self.made_batches = 0
             self.last_batch_time = time.time()
 
-            nb_sample = self.params['nb_sample'] #training samples total
-            nb_epoch = self.params['nb_epoch'] #training samples total
-            batch_size = self.current['batch_size'] #normal batch size
-            nb_batches = nb_sample / batch_size #normal nb batches
+            nb_sample = self.params['nb_sample']  # training samples total
+            batch_size = self.current['batch_size']  # normal batch size
+            nb_batches = nb_sample / batch_size  # normal nb batches
 
             self.current['batchesPerSecond'] = self.batches_per_second
             self.current['itemsPerSecond'] = self.batches_per_second * current_batch_size
 
-            epochs_per_second = self.batches_per_second / nb_batches #all batches
+            epochs_per_second = self.batches_per_second / nb_batches  # all batches
             self.current['epochsPerSecond'] = epochs_per_second
 
-            made_batches = (self.current['epoch'] * nb_batches) + current_batch
-            total_batches = nb_epoch * nb_batches
-            needed_batches = total_batches - made_batches
-            seconds_per_batch = 1 / self.batches_per_second
-
-            self.current['eta'] = seconds_per_batch * needed_batches
-
             self.current['currentBatch'] = current_batch
             self.current['currentBatchSize'] = current_batch_size
             elapsed = time.time() - self.start_time
             self.current['elapsed'] = elapsed
 
-            self.trainer.set_job_info('current', self.current)
+            self.trainer.set_job_info('itemsPerSecond', self.batches_per_second * current_batch_size)
+            self.trainer.set_job_info('currentBatch', current_batch)
+            self.trainer.set_job_info('currentBatchSize', batch_size)
+            self.trainer.set_job_info('nb_batches', nb_batches)
 
     def write(self, line):
         self.general_logger.write(line)
@@ -141,31 +156,11 @@ class KerasLogger(Callback):
         self.filter_invalid_json_values(log)
 
         log['created'] = time.time()
-        log['epoch'] = epoch+1
+        log['epoch'] = epoch + 1
         if 'loss' not in log and len(self.validation_per_batch) > 0:
             log['loss'] = sum(self.validation_per_batch) / float(len(self.validation_per_batch))
 
-        self.validation_per_batch = []
-        log['validation_accuracy'] = {}
-        log['validation_loss'] = {}
-        log['training_loss'] = {}
-        log['training_accuracy'] = {}
-
-        elapsed = time.time() - self.start_time
-
-        total_loss = 0
-        total_accuracy = 0
-
-        for layer in self.model.output_layers:
-            #todo, this is not very generic
-            log['validation_loss'][layer.name] = log['val_loss'] #outs[0]
-            log['validation_accuracy'][layer.name] = log['val_acc'] #outs[1]
-
-            log['training_loss'][layer.name] = log['loss'] #outs[0]
-            log['training_accuracy'][layer.name] = log['acc'] #outs[1]
-
-            total_loss += log['val_loss']
-            total_accuracy += log['val_acc']
+        total_accuracy = log.get('val_acc', 0)
 
         if total_accuracy > self.best_total_accuracy:
             self.best_total_accuracy = total_accuracy
@@ -174,40 +169,27 @@ class KerasLogger(Callback):
 
         self.model.save_weights(self.filepath_latest, overwrite=True)
 
-        if total_accuracy < self.worst_total_accuracy:
-            self.worst_total_accuracy = total_accuracy
+        self.loss_channel.send(log['epoch'], log.get('loss', 0), log.get('val_loss', 0))
+        self.accuracy_channel.send(log['epoch'], total_accuracy*100)
 
-        self.current['totalValidationLoss'] = total_loss
-        self.current['totalValidationAccuracy'] = total_accuracy
-        self.current['totalValidationAccuracyBest'] = self.best_total_accuracy
-        self.current['totalValidationAccuracyWorst'] = self.worst_total_accuracy
-        self.current['totalValidationAccuracyBestEpoch'] = self.best_epoch
-
-        self.current['totalTrainingLoss'] = log['loss']
-        self.current['elapsed'] = elapsed
-        self.current['epoch'] = log['epoch']
-
-        self.filter_invalid_json_values(self.current)
-
-        self.trainer.set_job_info('current', self.current)
+        self.job_backend.progress(log['epoch'], self.params['nb_epoch'])
 
         if self.log_epoch:
-            line = "Epoch %d: loss=%f, acc=%f, val_loss=%f, val_acc=%f\n" % (log['epoch'], log['loss'], log.get('acc'), log['val_loss'], log.get('val_acc'), )
+            line = "Epoch %d: loss=%f, acc=%f, val_loss=%f, val_acc=%f\n" % (
+                log['epoch'], log['loss'], log.get('acc', 0), log.get('val_loss', 0), total_accuracy, )
             self.general_logger.write(line)
 
-        self.backend.job_add_status('epoch', log)
-
         if self.job_model.job['insights']:
-            #Todo, support multiple inputs
+            # Todo, support multiple inputs
             first_input_layer = self.model.input_layers[0]
 
-            if first_input_layer != None:
+            if first_input_layer is not None:
 
                 images = self.build_insight_images()
                 # build confusion matrix
                 confusion_matrix = self.build_confusion_matrix() if self.confusion_matrix else None
 
-                self.backend.job_add_insight({'epoch': log['epoch'], 'confusionMatrix': confusion_matrix}, images)
+                self.job_backend.job_add_insight(log['epoch'], images, confusion_matrix)
 
     def is_image_shape(self, x):
         if len(x.shape) != 3 and len(x.shape) != 2:
@@ -234,18 +216,22 @@ class KerasLogger(Callback):
                 for layer in self.model.input_layers:
                     X = self.trainer.data_train['x'][layer.name]
                     if self.trainer.is_generator(X):
-                        batch_x, batch_y = X.next()
+                        batch_x, batch_y = next(X)
                         input_data_x.append([batch_x[0]])
                     else:
                         input_data_x.append([X[0]])
             else:
                 input_data_x = []
-                for X in self.trainer.data_train['x']:
-                    if self.trainer.is_generator(X):
-                        batch_x, batch_y = X.next()
-                        input_data_x.append([batch_x[0]])
-                    else:
-                        input_data_x.append([X[0]])
+                if self.trainer.is_generator(self.trainer.data_train['x']):
+                    batch_x, batch_y = next(self.trainer.data_train['x'])
+                    input_data_x.append([batch_x[0]])
+                else:
+                    for X in self.trainer.data_train['x']:
+                        if self.trainer.is_generator(X):
+                            batch_x, batch_y = next(X)
+                            input_data_x.append([batch_x[0]])
+                        else:
+                            input_data_x.append([X[0]])
 
         return input_data_x
 
@@ -260,11 +246,7 @@ class KerasLogger(Callback):
                 x = np.squeeze(self.insights_x[i])
                 if self.is_image_shape(x):
                     image = self.make_image(x)
-                    images.append({
-                        'id': layer.name,
-                        'title': layer.name,
-                        'image': self.to_base64(image)
-                    })
+                    images.append(JobImage(layer.name, image))
         except:
             pass
 
@@ -279,56 +261,51 @@ class KerasLogger(Callback):
         for layer in self.model.layers:
             if isinstance(layer, keras.layers.convolutional.Convolution2D) or isinstance(layer, keras.layers.convolutional.MaxPooling2D):
 
-                fn = K.function(inputs, [layer.output])
+                fn = K.function(inputs, self.get_layout_output_tensors(layer))
                 Y = fn(input_data_x_sample)[0]
 
                 data = np.squeeze(Y)
-                # print("Layer Activations " + layer.name)
                 image = PIL.Image.fromarray(get_layer_vis_square(data))
 
-                images.append({
-                    'id': layer.name,
-                    'type': 'convolution',
-                    'title': layer.name,
-                    'image': self.to_base64(image)
-                })
+                images.append(JobImage(layer.name, image))
 
                 if hasattr(layer, 'W') and layer.W:
-                    # print("Layer Weights " + layer.name)
-                    data = layer.W.get_value()
+                    data = layer.get_weights()[0]
                     image = PIL.Image.fromarray(get_layer_vis_square(data))
-                    images.append({
-                        'id': layer.name+'_weights',
-                        'type': 'convolution',
-                        'title': layer.name + ' weights',
-                        'image': self.to_base64(image)
-                    })
-
+                    images.append(JobImage(layer.name + '_weights', image, layer.name + ' weights'))
 
             if isinstance(layer, keras.layers.Dense):
 
-                fn = K.function(inputs, [layer.output])
+                fn = K.function(inputs, self.get_layout_output_tensors(layer))
                 Y = fn(input_data_x_sample)[0]
+                Y = np.squeeze(Y)
+
+                if Y.size == 1:
+                    Y = np.array([Y])
 
                 node = self.job_model.get_model_node(layer.name)
                 if node and node['activationFunction'] == 'softmax':
-                    image = self.make_image_from_dense_softmax(np.squeeze(Y))
+                    image = self.make_image_from_dense_softmax(Y)
                 else:
-                    image = self.make_image_from_dense(np.squeeze(Y))
+                    image = self.make_image_from_dense(Y)
 
-                images.append({
-                    'id': layer.name,
-                    'type': 'dense',
-                    'title': layer.name,
-                    'image': self.to_base64(image)
-                })
+                images.append(JobImage(layer.name, image))
 
         return images
 
+    def get_layout_output_tensors(self, layer):
+        outputs = []
+
+        for idx, node in enumerate(layer.inbound_nodes):
+            outputs.append(layer.get_output_at(idx))
+
+        return outputs
+
     def build_confusion_matrix(self):
         confusion_matrix = {}
 
-        model_has_validation_data = hasattr(self.model, 'validation_data') and self.model.validation_data
+        model_has_validation_data = hasattr(
+            self.model, 'validation_data') and self.model.validation_data
         if not model_has_validation_data and not self.trainer.data_validation:
             return confusion_matrix
 
@@ -350,26 +327,27 @@ class KerasLogger(Callback):
 
             input_data_y = np.squeeze(self.model.validation_data[len(self.model.input_layers)])
         else:
-            #model does not have validation_data attribute, which is the case when a generator is given
+            # model does not have validation_data attribute, which is the case when a
+            # generator is given
             if self.trainer.is_generator(self.trainer.data_validation):
                 input_data_x = self.trainer.data_validation
             else:
-                #it's probably struct of AETROS code generation
+                # it's probably struct of AETROS code generation
                 if 'x' in self.trainer.data_validation:
                     if self.trainer.is_generator(self.trainer.data_validation['x']):
                         input_data_x = self.trainer.data_validation['x']
                     elif isinstance(self.trainer.data_validation['x'], dict):
-                        for k, X in self.trainer.data_validation['x'].iteritems():
+                        for k, X in six.iteritems(self.trainer.data_validation['x']):
                             input_data_x = X
 
                         if not self.trainer.is_generator(input_data_x):
-                            for k, Y in self.trainer.data_validation['y'].iteritems():
+                            for k, Y in six.iteritems(self.trainer.data_validation['y']):
                                 input_data_y = Y
 
                     elif isinstance(self.trainer.data_validation['x'], list):
                         input_data_x = self.trainer.data_validation['x'][0]
                         if not self.trainer.is_generator(input_data_x):
-                                input_data_y = self.trainer.data_validation['y'][0]
+                            input_data_y = self.trainer.data_validation['y'][0]
 
         if input_data_x is None:
             return confusion_matrix
@@ -380,7 +358,7 @@ class KerasLogger(Callback):
             processed_samples = 0
 
             while processed_samples < self.trainer.nb_val_samples:
-                generator_output = input_data_x.next()
+                generator_output = next(input_data_x)
                 if len(generator_output) == 2:
                     x, y = generator_output
                     sample_weight = None
@@ -413,7 +391,8 @@ class KerasLogger(Callback):
                     pass
 
         else:
-            prediction = self.model.predict(input_data_x, batch_size=self.job_model.get_batch_size())
+            prediction = self.model.predict(
+                input_data_x, batch_size=self.job_model.get_batch_size())
             predicted_classes = prediction.argmax(axis=-1)
             expected_classes = np.array(input_data_y).argmax(axis=-1)
 
@@ -429,17 +408,12 @@ class KerasLogger(Callback):
         return confusion_matrix
 
     def filter_invalid_json_values(self, dict):
-        for k,v in dict.iteritems():
+        for k, v in six.iteritems(dict):
             if isinstance(v, (np.ndarray, np.generic)):
                 dict[k] = v.tolist()
             if math.isnan(v) or math.isinf(v):
                 dict[k] = -1
 
-    def to_base64(self, image):
-        buffer = cStringIO.StringIO()
-        image.save(buffer, format="JPEG", optimize=True, quality=80)
-        return base64.b64encode(buffer.getvalue())
-
     def make_image(self, data):
         from keras.preprocessing.image import array_to_img
         try:
@@ -455,7 +429,7 @@ class KerasLogger(Callback):
         from aetros.utils import array_to_img
 
         img = array_to_img(neurons.reshape((1, len(neurons), 1)))
-        img = img.resize((9, len(neurons)*8))
+        img = img.resize((9, len(neurons) * 8))
 
         return img
 
@@ -463,12 +437,12 @@ class KerasLogger(Callback):
         from aetros.utils import array_to_img
         cols = int(math.ceil(math.sqrt(len(neurons))))
 
-        even_length = cols*cols
+        even_length = cols * cols
         diff = even_length - len(neurons)
         if diff > 0:
             neurons = np.append(neurons, np.zeros(diff, dtype=neurons.dtype))
 
         img = array_to_img(neurons.reshape((1, cols, cols)))
-        img = img.resize((cols*8, cols*8))
+        img = img.resize((cols * 8, cols * 8))
 
         return img
